#include <stdlib.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#define THROW_BIT PB5
#define DICE_BIT PB4

#define diceNumber() (PINB & _BV(DICE_BIT) ? 2 : 1)
#define isThrown() !(PINB & _BV(THROW_BIT))
#define pcintEnabled(enabled) \
	if (enabled) { PCMSK = _BV(THROW_BIT) | _BV(DICE_BIT); } \
	else { PCMSK = 0; }

uint8_t num2dice(uint8_t num);
void multiplexLeds();
void bleep();
void throwDice();
void changeDiceNumber();

// global state of displayed LEDs in format returned by num2dice()
volatile uint8_t g_diceRight = 0;
volatile uint8_t g_diceLeft = 0;
volatile uint8_t g_thrown = 0;
volatile uint8_t g_diceNum = 0;
volatile uint8_t g_diceNumChanged = 0;

int main(void) {
	// switches as inputs.
	DDRB = ~(_BV(THROW_BIT) | _BV(DICE_BIT));
	// pull-ups for input, set bit for outputs
	PORTB = 0xFF;

	// Port D is output-only, 4 higher bits are not used
	DDRD = 0xFF;
	PORTD = 0xFF;

	// Timer0 with prescaler 1:1
	TCCR0B = _BV(CS00);
	// Overflow interrupt for Timer0
	TIMSK = _BV(TOIE0);
	// Enable PCINT interrupt on switches
	pcintEnabled(1);
	// PINB change interruption PCINT
	GIMSK = _BV(PCIE);

	// Enable interrupts in SREG
	sei();

	g_diceNum = diceNumber();
	g_diceNumChanged = 1;
	g_thrown = 1;

	while (1) {
		if (g_diceNumChanged) {
			g_diceNumChanged = 0;
			changeDiceNumber();
		}
		if (g_thrown) {
			g_thrown = 0;
			throwDice();
		}
	}
}

ISR(PCINT_vect)
{
	uint8_t diceNum = diceNumber();
	if (g_diceNum != diceNum) {
		g_diceNum = diceNum;
		g_diceNumChanged = 1;
	} else if (isThrown()) {
		g_thrown = 1;
	}
}

ISR(TIMER0_OVF_vect)
{
	multiplexLeds();
}

/**
 * Make a short cracky sound.
 */
void bleep() {
	for (uint8_t i = 0; i < 50; i++) {
		PORTD ^= _BV(PB4);
		_delay_us(10);
	}
}

/**
 * Multiplexing LED display. Since LED with different colors have different brightness,
 * multiplexed columns light on with different times (longer for less bright diodes).
 */
#define COLUMNS 4
void multiplexLeds() {
	// number of cycles spent for each column
	static uint8_t columnCycles[COLUMNS] = { 4, 4, 1, 1 };
	static uint8_t column = 0;
	static uint8_t columnCycle = 0;
	if (--columnCycle <= 0) {
		if (++column >= COLUMNS) {
			column = 0;
		}
		columnCycle = columnCycles[column];
		PORTB |= 0x0F;
		PORTD |= 0x0F;
		switch (column) {
		case 0:
			// lower 4 bits of left dice
			PORTD &= ~g_diceLeft | 0xF0;
			break;
		case 1:
			// higher 4 bits of left dice (shifted to lower 4 bits of port D)
			PORTD &= (~(g_diceLeft >> 4)) | 0xF0;
			break;
		case 2:
			// lower 4 bits of right dice
			PORTD &= ~g_diceRight | 0xF0;
			break;
		case 3:
			// higher 4 bits of right dice (shifted to lower 4 bits of port D)
			PORTD &= (~(g_diceRight >> 4)) | 0xF0;
			break;
		}
		// light on column
		PORTB &= ~_BV(column) | 0xF0;
	}
}

/**
 * Converts number for 1 to 6 into bit flags of LED to light on.
 * Single dice in hardware is represented by 4 rows and 2 columns of LEDS;
 * middle LED is mapped from 2nd column - find details below (row, column)
 * notation zero-indexed:
 *
 *  LED3(2,0)             LED6(2,1)
 *
 *  LED2(1,0)  LED7(3,1)  LED5(1,1)
 *
 *  LED1(0,0)             LED4(0,1)
 *
 * Position (3,0) is unused. Lower 4 bits of returned byte are mapped into
 * first column, while higher 4 bits to second column. Bit 4th is unused as
 * is mapped to position (3,0).
 *
 * Example: number "3" on dice are LEDs light diagonally: (0,0) (3,1) and (2,1)
 * which turns into returned value 0b11000001. Similarly "6" has all LEDs light
 * except for middle (3,1) which means 0b01110111 (4th bit is always zero as
 * unused).
 */
uint8_t num2dice(uint8_t num) {
	uint8_t ret = 0;
	switch (num) {
	case 1:
		ret = 0b10000000;
		break;
	case 2:
		ret = 0b01000001;
		break;
	case 3:
		ret = 0b11000001;
		break;
	case 4:
		ret = 0b01010101;
		break;
	case 5:
		ret = 0b11010101;
		break;
	case 6:
		ret = 0b01110111;
		break;
	}
	return ret;
}

void throwDice() {
	pcintEnabled(0);
	srand(TCNT0);
	// choose numbers couple times as if dice were thrown
	// before each number increase delay (dice is slowing down ;)
	for (uint8_t i = 0; i < 5; i++) {
		if (g_diceNum > 1) {
			g_diceLeft = num2dice(rand() % 6 + 1);
		} else {
			g_diceLeft = 0;
		}
		g_diceRight = num2dice(rand() % 6 + 1);
		bleep();
		_delay_ms(15);
		for (uint8_t j = 0; j < i; j++) {
			_delay_ms(8);
		}
	}
	pcintEnabled(1);
}

void changeDiceNumber() {
	pcintEnabled(0);
	// circle dice leds three times to notify user
	// about number of dice used
	uint8_t sequence[6] = { 0x01, 0x10, 0x20, 0x40, 0x04, 0x02 };
	uint8_t backupRight = g_diceRight;
	for (uint8_t i = 0; i < 3*6; i++) {
		uint8_t s = sequence[i%6];
		if (g_diceNum > 1) {
			g_diceLeft = s;
		} else {
			g_diceLeft = 0;
		}
		g_diceRight = s;
		_delay_ms(4);
		bleep();
	}
	if (g_diceNum > 1) {
		g_diceLeft = g_diceRight = backupRight;
	} else {
		g_diceLeft = 0;
		g_diceRight = backupRight;
	}
	pcintEnabled(1);
}
